import numpy as np

def in_dis_vec(data, U, x, curr_val):
    row1 = data.shape[0]
    row, col = U.shape
    diff_dec = (data[x, col-1] != U[:, col-1])
    curr = np.ones(row, dtype=bool)
    in_val = np.sum(np.max(U[diff_dec, :-1] != data[x, :-1], axis=1))
    in_val = (in_val + curr_val*row1) / (row1 + row)
    return in_val

def incre_discernible_score(data, U, sam_left1, sam_left2, curr_ds_vector, add_ds_vector):
    NL1 = len(sam_left1)
    NL2 = len(sam_left2)
    in_ds_vec = np.zeros(NL1)
    in_add_ds_vec = np.zeros(NL2)
    
    for i in range(NL1):
        in_ds_vec[i] = in_dis_vec(data, U, sam_left1[i], curr_ds_vector[i])
    
    for i in range(NL2):
        in_add_ds_vec[i] = in_dis_vec(U, data, sam_left2[i], add_ds_vector[i])
    
    in_ds = np.sum(in_ds_vec) + np.sum(in_add_ds_vec)
    in_ds /= (NL1 + NL2)
    
    return in_ds, in_ds_vec, in_add_ds_vec



import numpy as np
import time
import numpy as np

def discerbile_vec(data, x):
    diff_dec = (data[x, -1] != data[:, -1])
    if diff_dec.any():
        data_filtered = data[diff_dec, :-1] - data[x, :-1]
        # data_filtered empty? 
        if data_filtered.size > 0:
            inVal = np.sum(np.max(data_filtered, axis=1))
        else:
            inVal = 0
    else:
        inVal = 0
    return inVal

def discernible_score(data, sample_left):
    NL = len(sample_left)
    ds_vector = np.zeros(NL)
    for i in range(NL):
        ds_vector[i] = discerbile_vec(data, sample_left[i])
    DS = np.sum(ds_vector) / NL if NL > 0 else 0
    return DS, ds_vector


def fast_feature_selection_ds(data):
    start_time = time.time()

    row, col = data.shape
    fea_slt = np.array([], dtype=int)  
    fea_left = np.arange(col - 1)  
    sam_left = np.arange(row)  

    # Obtaining discernible score for all samples
    _, ds_vector = discernible_score(data, sam_left)
    DSSamLeft = 0
    fea_redun = []  # To store redundant features
    sam_delete = []  # To store deleted samples

    while len(sam_left) > 0 and len(fea_left) > 0:
        tempDS = np.zeros(len(fea_left))
        for i, fea in enumerate(fea_left):
            selected_features = np.append(fea_slt, fea)  
            _, temp_ds_vec = discernible_score(data[:, np.append(selected_features, col-1)], sam_left)
            tempDS[i] = np.sum(temp_ds_vec) / len(sam_left)  # Calculating the average discernible score

        maxNum, maxInd = np.max(tempDS), np.argmax(tempDS)  
        fea_slt = np.append(fea_slt, fea_left[maxInd])  
        fea_left = np.delete(fea_left, maxInd)  # Removing the selected feature from the list of remaining features

        tempDS = np.delete(tempDS, maxInd)  # Removing the discernible score of the selected feature
        filter_mask = tempDS - DSSamLeft > 0.001  
        fea_left = fea_left[filter_mask]  

        fea_redun.append(len(fea_left))  # Updating the number of redundant features
        dsVector_diff = ds_vector[sam_left] - temp_ds_vec  # Calculating the difference in discernible score
        NDelSam = np.where(dsVector_diff <= 0.001)[0]
        sam_left = np.delete(sam_left, NDelSam)  # Removing redundant samples
        sam_delete.append(len(NDelSam))  # Updating the number of deleted samples

        DSSamLeft = maxNum  # Updating the total discernible score

    time_elapsed = time.time() - start_time  

    return fea_slt.tolist(), ds_vector, fea_redun, sam_delete, time_elapsed



import numpy as np
import time
from HighdimensionalFeat.increFeatureSelectionDS import incre_feature_selection_ds


def incre_fea_slt_ds_filter_sam(data, U, curr_ds_vector, fea_slt):
    start_time = time.time()

    row1, _ = data.shape
    row2 = U.shape[0]
    combined_data = np.vstack((data, U))
    _, use_ds = np.unique(combined_data, axis=0, return_index=True)
    use_ds1 = use_ds[use_ds < row1]
    use_ds2 = use_ds[use_ds >= row1] - row1
    unuse_ds = row1 + row2 - len(use_ds)
    NRF = []
    NRS = []

    if len(use_ds2) == 0 or len(use_ds1) == 0:
        in_fea_slt = fea_slt
    else:
        in_fea_slt, NRF, NRS, _ = incre_feature_selection_ds(data[use_ds1, :], U[use_ds2, :], curr_ds_vector[use_ds1], fea_slt)

    time_elapsed = time.time() - start_time
    return in_fea_slt, unuse_ds, NRF, NRS, time_elapsed



import numpy as np

def partition_fold(data, k, max_iterations=100):
    """
    Function to partition the dataset into k parts using the Grey Wolf Optimizer.
    
    Parameters:
    data (np.array): The dataset to be divided.
    k (int): The desired number of parts.
    max_iterations (int): The number of iterations for the Grey Wolf Optimizer.
    
    Returns:
    list of np.array: A list containing k parts of the dataset.
    """
    def objective_function(indices):
        # Reshape indices into k parts
        parts = [indices[i::k] for i in range(k)]
        # Calculate the objective (e.g., variance of parts)
        variances = [np.var(data[part], axis=0).sum() for part in parts]
        return np.mean(variances)
    
    def initialize_positions(n_agents, dim):
        return np.random.randint(0, dim, (n_agents, dim))

    def gwo_optimize(dim, n_agents, max_iterations):
        alpha_pos = np.zeros(dim)
        alpha_score = float('inf')

        beta_pos = np.zeros(dim)
        beta_score = float('inf')

        delta_pos = np.zeros(dim)
        delta_score = float('inf')

        positions = initialize_positions(n_agents, dim)

        for iteration in range(max_iterations):
            for i in range(n_agents):
                fitness = objective_function(positions[i])

                if fitness < alpha_score:
                    alpha_score = fitness
                    alpha_pos = positions[i].copy()

                elif fitness < beta_score:
                    beta_score = fitness
                    beta_pos = positions[i].copy()

                elif fitness < delta_score:
                    delta_score = fitness
                    delta_pos = positions[i].copy()

            a = 2 - iteration * (2 / max_iterations)

            for i in range(n_agents):
                for j in range(dim):
                    r1, r2 = np.random.rand(2)
                    A1 = 2 * a * r1 - a
                    C1 = 2 * r2
                    D_alpha = abs(C1 * alpha_pos[j] - positions[i][j])
                    X1 = alpha_pos[j] - A1 * D_alpha

                    r1, r2 = np.random.rand(2)
                    A2 = 2 * a * r1 - a
                    C2 = 2 * r2
                    D_beta = abs(C2 * beta_pos[j] - positions[i][j])
                    X2 = beta_pos[j] - A2 * D_beta

                    r1, r2 = np.random.rand(2)
                    A3 = 2 * a * r1 - a
                    C3 = 2 * r2
                    D_delta = abs(C3 * delta_pos[j] - positions[i][j])
                    X3 = delta_pos[j] - A3 * D_delta

                    # Weights based on long-tail distribution (alpha > beta > delta)
                    weights = np.abs(np.random.pareto(a=2.5, size=3))  
                    weights /= np.sum(weights) 

                 
                    new_position = weights[0] * X1 + weights[1] * X2 + weights[2] * X3
                    positions[i][j] = np.clip(new_position, 0, dim - 1)  

        return alpha_pos


    num_samples = data.shape[0]
    num_agents = 20  # You can adjust this based on the problem size
    optimal_indices = gwo_optimize(num_samples, num_agents, max_iterations)
    optimal_indices = optimal_indices.astype(int)

    # Partition the data using the optimized indices
    parts = [data[optimal_indices[i::k], :] for i in range(k)]
    
    return parts



import numpy as np
import time
from discernibleScore import discernible_score
from increDiscernibleScore import incre_discernible_score

def incre_feature_selection_ds(data, U, curr_ds_vector, fea_slt):
    start_time = time.time()

    row1, col = data.shape
    row2 = U.shape[0]
    sam_left1 = np.arange(row1)
    sam_left2 = np.arange(row2)
    fea_left = np.setdiff1d(np.arange(col - 1), fea_slt)  # Removing already selected features
    NRF = []
    NRS = []

    _, add_ds_vector = discernible_score(U, sam_left2)
    in_ds, in_ds_vec, in_add_ds_vec = incre_discernible_score(data, U, sam_left1, sam_left2, curr_ds_vector, add_ds_vector)
    _, add_fea_ds_vec = discernible_score(U[:, np.append(fea_slt, col-1)], sam_left2)
    in_fea_ds, _, _ = incre_discernible_score(data[:, np.append(fea_slt, col-1)], U[:, np.append(fea_slt, col-1)], sam_left1, sam_left2, curr_ds_vector, add_fea_ds_vec)

    if in_ds - in_fea_ds <= 0.001:
        in_fea_slt = fea_slt
    else:
        add_fea = []
        left_sam_ds = 0
        num_curr = 1
        while len(fea_left) > 0 and (len(sam_left1) + len(sam_left2)) > 0:
            temp_in_ds = np.zeros(len(fea_left))
            for i, fea in enumerate(fea_left):
                selected_features = np.append(fea_slt, fea)
                _, fea_ds_vec = discernible_score(U[:, selected_features], sam_left2)
                _, ds_vec = discernible_score(data[:, selected_features], sam_left1)
                temp_in_ds[i], _, _ = incre_discernible_score(data[:, selected_features], U[:, selected_features], sam_left1, sam_left2, ds_vec, fea_ds_vec)
            max_num, max_ind = np.max(temp_in_ds), np.argmax(temp_in_ds)
            add_fea = np.append(add_fea, fea_left[max_ind])
            mask = temp_in_ds - left_sam_ds < 0.001
            
            # Update the mask to ensure its size matches fea_left after removal
            mask = np.delete(mask, max_ind)
            fea_left = np.delete(fea_left, max_ind)
            fea_left = fea_left[mask]
            NRF.append(len(fea_left))
            del_sam1 = in_ds_vec - temp_in_ds[max_ind] <= 0.001
            del_sam2 = in_add_ds_vec - temp_in_ds[max_ind] <= 0.001
            sam_left1 = np.delete(sam_left1, np.where(del_sam1)[0])
            sam_left2 = np.delete(sam_left2, np.where(del_sam2)[0])
            NRS.append(np.sum(del_sam1) + np.sum(del_sam2))
            left_sam_ds = max_num
            num_curr += 1

        in_fea_slt = np.unique(np.append(fea_slt, add_fea))

    time_elapsed = time.time() - start_time
    return in_fea_slt, NRF, NRS, time_elapsed



from setuptools import setup, find_packages

setup(
    name="HDDFeat",  
    version="0.1.0",  #
    description="A library for feature selection in high-dimensional datasets",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Said Al Afghani Edsa, Khamron Sunat, P.hD",
    author_email="saidalafghani.dumai@gmail.com",
    url="https://github.com/saiddddd/HDDFeat",  
    packages=find_packages(),
    install_requires=[],  
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import numpy as np

def discerbile_vec(data, x):
    diff_dec = (data[x, -1] != data[:, -1])
    if diff_dec.any():
        data_filtered = data[diff_dec, :-1] - data[x, :-1]
        # data_filtered empty? 
        if data_filtered.size > 0:
            inVal = np.sum(np.max(data_filtered, axis=1))
        else:
            inVal = 0
    else:
        inVal = 0
    return inVal

def discernible_score(data, sample_left):
    NL = len(sample_left)
    ds_vector = np.zeros(NL)
    for i in range(NL):
        ds_vector[i] = discerbile_vec(data, sample_left[i])
    DS = np.sum(ds_vector) / NL if NL > 0 else 0
    return DS, ds_vector
